<?php

  /**
  * Exception generated when SQL query execution fails
  */
  class EQueryException extends EException {
    protected $lastQuery;
    /**
    * Returns last executed query 
    *
    * @return string
    */
    public function getLastQuery() {
      return $this->lastQuery;
    }
    
    public function __construct($messageOrCode = null, $codeOrArgs = 0) {
      $this->lastQuery = array_pop($codeOrArgs);
      parent::__construct($messageOrCode,$codeOrArgs);           
    }    
  } 

  /**
  *
  */
  class TField extends TObject {
    /** @var TDataSet Dataset that this field belongs to */
    public $dataset;
    /** @var mixed Original value obtained from the query */
    public $oldValue;
    /** @var TCustomValue Current value of the field */
    public $value;
    /** @var string Field name */
    public $name;
    /** @var int Field index */
    public $index;
    /** @var int Width of the field */
    public $length;
    /** @var int Number of decimals */
    public $decimals;
    /** @var string Table that this field belongs to */
    public $table;
    /** @var string Original table name, if an alias has been used */
    public $orgTable;
    /** @var string Original field name, if an alias has been used */
    public $orgName;
    /** @var string Default value as a string */
    public $defaultValue;
    /** @var string Road data type */
    public $dataType;
    /** @var mixed Data type defined in database server */
    public $nativeDataType;
    /** @var bool Is the field part of the primary key? */
    public $primaryKey = false;
    /** @var bool Does the server automatically assign value to this field? */
    public $autoValue = false;
    /** @var bool Does the server disallow NULL value? */
    public $notNull = false;
    /** @var bool Should the field value be unique? */
    public $unique = false;    
    /** @var bool */
    private $changed;
    /** 
    * Resets "changed" flag to false
    *
    * @return void
    */
    protected function resetChanged()  {
      $this->changed = false;    
    }

    public function __clone ( )  {
      $this->cloneProperty('value');
      $this->value->parent = $this;
    }    
    public function valueChanged(TCustomValue $value) {
      $this->changed = true;
    }
    /**
    * Set data type property and create coresponding TCustomValue object
    *
    * @param int Data type
    * @return void
    */
    public function setDataType($dataType)  {
      global $__roadDataTypeValueMatrix;

      if(($dataType == TYPE_UNKNOWN) || !isset($__roadDataTypeValueMatrix[$dataType])) {
        $dataType = TYPE_UNKNOWN;
        triggerError(E_UNKNOWN_DATA_TYPE_FOR_FIELD,array($this->name),E_USER_WARNING);
      }
      $this->dataType = $dataType;

      $this->value = new $__roadDataTypeValueMatrix[$dataType]($this);
      $this->value->strict = true;
    }
    /**
    * Set the old and new value
    *
    * @return void
    */
    public function setOldValue($value) {
      $this->oldValue = $value;
      if($value === null) {
        $this->value->clear();
      } else {
        if(!$this->value->setAsSql($value)) {
          if($this->dataset->connection->debug) {
            throw new EException(E_INCOMPATIBLE_VALUE_FOR_FIELD,array($value,$this->name));
          } else {
            triggerError(E_INCOMPATIBLE_VALUE_FOR_FIELD,array($value,$this->name),E_USER_WARNING);
          }
        }
      }
      $this->resetChanged();
    }
    /**
    * Has the value changed since it has been loaded from the DB
    *
    * @return boolean
    */
    public function hasChanged()  {
      return $this->changed;
/*      $tmp = $this->value->getRaw();
      if($this->oldValue === null) {
        $this->value->clear();
      } else {
        $this->value->setAsSql($this->oldValue);
      }
      $res = ($tmp !== $this->value->getRaw());
      $this->value->setRaw($tmp);
      return $res;*/
    }
    /**
    * Is the field empty?
    *
    * @return bool
    */
    public function isNull() {
      return $this->value->isNull();
    }
    public function __construct(TDataSet $dataset) {
      $this->dataset = $dataset;
    }
    /**
    * Returns old value formated for SQL queries
    *
    * @return void
    */
    public function getOldValueAsSqlQuoted() {
      $changed = $this->changed;
      $tmp = $this->value->getRaw();
      
      if($this->oldValue === null) {
        $this->value->clear();
      } else {
        $this->value->setRaw($this->oldValue);
      }      

      $res = $this->value->getAsSqlQuoted();
      $this->value->setRaw($tmp);
//      $this->changed = $changed;
      return $res;
    }    
  }

  /**
  *
  */
  abstract class TDataSet extends TObject {
    private $inserting = false;    
    /** @var TCustomConnection Connection that created this dataset */
    public $connection;
    /** @var TField[] Fields by name */
    public $fields;
    /** @var TField[] Fields by index */
    public $fieldsByIndex;
    /** @var string Table name*/
    public $table;
    protected $freeOnRelease = true;
    private $lastRow = array();    
    protected $fieldNames = array();
    private $query;

    /**                    
    * Set value of selected field formated for output
    *
    * @return bool
    */
    public function setAsDisplay($fieldNameOrIndex,$value)  {
      $f = $this->getField($fieldNameOrIndex);
      return $f->value->setAsDisplay($value);    
    }
        
    /**                    
    * Return value of selected field formated for output
    *
    * @return string
    */
    public function getAsDisplay($fieldNameOrIndex)  {
      $f = $this->getField($fieldNameOrIndex);
      return $f->value->getAsDisplay();    
    }
    /**                    
    * Return true if selected field is not empty 
    *
    * @return bool
    */    
    public function hasValue($fieldNameOrIndex)  {
      $f = $this->getField($fieldNameOrIndex);
      return (!$f->value->isNull());
    }

    /**                    
    * Return value of selected field formated for input
    *
    * @return string
    */
    public function getAsInput($fieldNameOrIndex)  {
      $f = $this->getField($fieldNameOrIndex);
      return $f->value->getAsInput();    
    }
    /**                    
    * Set value of selected field formated for input
    *
    * @return bool
    */
    public function setAsInput($fieldNameOrIndex,$value)  {
      $f = $this->getField($fieldNameOrIndex);
      return $f->value->setAsInput($value);    
    }
    /**                    
    * Return value of selected field in raw PHP format
    *
    * @return mixed
    */
    public function getRaw($fieldNameOrIndex)  {
      $f = $this->getField($fieldNameOrIndex);
      return $f->value->getRaw();
    }


    /**                    
    * Set value of selected field in raw PHP format
    *
    * @return bool
    */
    public function setRaw($fieldNameOrIndex,$value)  {
      $f = $this->getField($fieldNameOrIndex);
      return $f->value->setRaw($value);
    }

    /**                    
    * Return value of selected field formated for SQL
    *
    * @return string
    */
    public function getAsSql($fieldNameOrIndex)  {
      $f = $this->getField($fieldNameOrIndex);
      return $f->value->getAsSql();    
    }
    /**                    
    * Set value of selected field formated for SQL
    *
    * @return bool
    */
    public function setAsSql($fieldNameOrIndex,$value)  {
      $f = $this->getField($fieldNameOrIndex);
      return $f->value->setAsSql($value);    
    }    
    protected function setQuery($query)  {
      $this->query = $query;
    }
    /**
    * Returns query that created this dataset
    *
    * @return string
    */
    public function getQuery()  {
      return $this->query;
    }
    /** @var int Current row index */
    protected $rowIndex = 0;
    public function __clone() {
      $this->freeOnRelease = false;
      $this->cloneProperty('fieldsByIndex');
      $this->fields = array();
      foreach($this->fieldsByIndex as $field) {
        $field->dataset = $this;
        $this->fields[$field->name] = $field;
      }      
    }    
    public function __construct(TCustomConnection $connection) {
      $this->connection = $connection;
    }
    public function cloneMe()  {      
      $this->freeOnRelease = false;
      return clone $this;
    }
    /**
    * Returns single string key uniquely identifiing current row
    *
    * @return string
    */
    public function getRowId() {
      $vals = array();
      foreach($this->fields as /** @var TField */ $field) {
        if($field->primaryKey) {
          $vals[] = urlencode($field->value->getAsSql());
        }
      }
      return implode('-',$vals);
    }
    /**
    * Parses row id obtained from getRowId()
    *
    * @return string[]
    */
    public function parseRowId($value) {
      if(!is_array($this->fields)) {
        $this->fetchFields();
      }
      $vals = explode('-',$value);
      $res = array();
      $index = 0;
      foreach($this->fields as /** @var TField */ $field) {
        if($field->primaryKey) {
          if(!isset($vals[$index])) {
            return false;
          }
          $res[$field->name] = urldecode($vals[$index]);
          $index ++;
        }
      }
      return $res;
    }
    /**
    * Delete current record
    *
    * @return bool
    */
    public function delete() {
      if($this->isInserting()) {
        $this->cancel();
        return true;
      }
      /** @var TMySqlImprovedConnection */      
      $db = $this->getDb();
      if($db->execute(
        'DELETE FROM '
        .$db->quoteFieldName($this->table)
        .' WHERE '
        .$this->getPrimaryKeyCondition())
      ) {
        if($db->getAffectedCount() < 1) {
          triggerError(E_RECORD_CANNOT_BE_DELETED,array(formatError(E_RECORD_NOT_FOUND)),E_USER_WARNING);
          return false;
        }
        return true;
      } else {
        throw new EException(E_RECORD_CANNOT_BE_DELETED,array(''));
      }
      
    }
    /**
    * Returns enumerated array of primary key fields
    *
    * @return TField[]
    */
    public function getKeyFields() {
      if(!is_array($this->fields)) {
        $this->fetchFields();
      }
      $res = array();
      foreach($this->fields as /** @var TField */ $field) {
        if($field->primaryKey) {
          $res[] = $field;
        }
      }
      return $res;
    }    
    /**
    * Return field by name or index (throws an exception if the field does not exists)
    *
    * @param mixed Field name or index 
    * @return TField
    */
    public function getField ($fieldNameOrIndex)  {
      if(!is_array($this->fields)) {
        $this->fetchFields();
      }    
      if(is_int($fieldNameOrIndex)) {
        if(!isset($this->fieldsByIndex[$fieldNameOrIndex])) {
          throw new EException(E_FIELD_NOT_FOUND,array($fieldNameOrIndex));
        }
        return $this->fieldsByIndex[$fieldNameOrIndex]; 
      } else {
        if(!isset($this->fields[$fieldNameOrIndex])) {
          throw new EException(E_FIELD_NOT_FOUND,array($fieldNameOrIndex));
        }
        return $this->fields[$fieldNameOrIndex];
      }
    }
    /**
    * Returns primary key query
    *
    * @return void
    */
    public function getPrimaryKeyCondition($separator = ' AND ')  {
      $res = '';
      foreach($this->fields as /** @var TField */ $field) {
        if($field->primaryKey) {
          if($res) {
            $res .= $sepparator;
          }
          $res .= $this->connection->quoteFieldName($field->name);
          if($field->oldValue === null) {
            $res .= ' IS NULL';
          } else {
            $res .= ' = '.$field->getOldValueAsSqlQuoted();
          }
        }
      }
      if($res) {
        return $res;
      } else {
        return false;
      }
    }
    /**
    * Is the dataset in insert mode
    *
    * @return bool
    */
    public function isInserting()  {
      return $this->inserting;
    }
    /**
    * Is the dataset in edit mode
    *
    * @return bool
    */
    public function isEditing()  {
      return !$this->inserting;
    }
    /**
    * Insert new row
    *
    * @return void
    */
    public function insert()  {
      if(!is_array($this->fields)) {
        $this->fetchFields();
      }
      $this->inserting = true;
      foreach($this->fields as $field) {      
        $field->setOldValue(null);
      }
    }
    /**
    * Returns true if the dataset contains no rows
    *
    * @return bool
    */
    public function isEmpty() {
      return ($this->getRowCount() == 0);
    }
    /**
    * Returns field index of a field specified by field name or index
    *
    * @return int Field index
    */
    public function fieldNameToIndex($fieldNameOrIndex) {
      if(is_int($fieldNameOrIndex)) {
        return $fieldNameOrIndex;
      }
      if(isset($this->fields[$fieldNameOrIndex])) {
        return $this->fields[$fieldNameOrIndex]; 
      } else {
        return -1;
      } 
    }
    abstract protected function doFetchField($index);
    
    private function loadFieldValues() {
      if(!$this->lastRow) {
        foreach($this->fields as $field) {      
          $field->setOldValue(null);
        }
      } else {
        foreach($this->fields as $field) {
          $field->setOldValue($this->lastRow[$field->name]);
        }
      }
    }
    /**
    * Load field definition from the result, create coresponding TField objects and move to next row
    *
    * @return bool
    */
    public function fetchFields()  {    
      if(!is_array($this->fields)) {
        for($i = 0; $i < $this->getFieldCount(); $i++) {
          $f = $this->doFetchField($i);
          $f->value->setDb($this->connection);
          if(is_null($this->table) && $f->primaryKey) {
            $this->table = $f->orgTable;
          }

          $this->fields[$f->name] = $f;
          $this->fieldsByIndex[$i] = $f;
          unset($f);
        }      
      }
      $this->lastRow = $this->fetchAssoc();

      $this->loadFieldValues();      
      if($this->lastRow) {
        return true;
      } else {
        return false;            
      }
    } 
    /**
    * Returns the row count
    *
    * @return int
    */
    abstract public function getRowCount();
    /**
    * Returns the field count
    *
    * @return int
    */
    abstract public function getFieldCount();
    abstract protected function doSeekTo(&$rowIndex);
    /**
    * Returns the current row index
    *
    * @return int
    */
    public function getRowIndex()  {
      return $this->rowIndex;
    }
    
    /**
    * Move the row cursor by offset
    *
    * @param int Offset
    * @return void
    */
    public function moveBy($offset)  {
      $offset = $this->getRowIndex() + $offset;
      return $this->seek($offset);
    }
    /**
    * Move to next row
    *
    * @return bool
    */
    public function next()  {
      return $this->moveBy(1);
    }
    /**
    * Move to previous row
    *
    * @return bool
    */
    public function prior()  {
      return $this->moveBy(-1);
    }
    /**
    * Seek to first row
    *
    * @return bool
    */
    public function first()  {
      return $this->seek(0);
    }
    /**
    * Seek to last row
    *
    * @return bool
    */
    public function last()  {
      return $this->seek($this->getRowCount()-1);
    }
    /**
    * Seek to desired position
    *
    * @param int Row index to seek to
    * @return bool
    */
    public function seek($rowIndex) {
      if(($rowIndex < 0) || ($rowIndex >= $this->getRowCount())) {
        return false;
      }
      if($this->doSeekTo($rowIndex)) {
        $this->rowIndex = $rowIndex;
        return true;
      }
      return false;
    }
    abstract protected function doFetchObject();
    /**
    * Fetch row values as an object and move to next row
    *
    * @return object
    */
    public function fetchObject() {
      $res = $this->doFetchObject();
      if($res) {
        $this->rowIndex ++;
      }
      return $res;     
    }

    abstract protected function doFetchArray();
    /**
    * Fetch row values as an enumerated array and move to next row
    *
    * @return mixed[]
    */
    public function fetchArray() {
      $res = $this->doFetchArray();
      if($res) {
        $this->rowIndex ++;
      }
      return $res;
    }
    abstract protected function doFetchAssoc();
    /**
    * Cancel changes made on the dataset
    *
    * @return void
    */
    public function cancel()  {
      if(!is_array($this->fields)) {
        throw EException(E_CALL_FETCH_FIELDS_OR_INSERT_FIRST);
      }    
      if($this->inserting) {
        $this->loadFieldValues();        
      } else {
        foreach($this->fields as $field) {
          $field->setOldValue($field->oldValue);
        }
      }
      $this->inserting = false;
    }
    /**
    * Post changes made on the dataset to database server 
    *
    * @return bool
    */
    public function post()  {
      if(!is_array($this->fields)) {
        throw EException(E_CALL_FETCH_FIELDS_OR_INSERT_FIRST);
      }
      if($this->isInserting()) {
        $names = '';
        $values = '';
        $res = false;
        foreach($this->fields as /** @var TField */ $field) {
          if($field->isNull()) {
            if($field->notNull && !$field->autoValue) {
              triggerError(E_FIELD_CANNOT_BE_NULL,array($field->name),E_USER_WARNING);
            }
            continue;
          }
          if($field->orgTable != $this->table) {
            triggerError(E_CANNOT_UPDATE_JOINED_TABLE_FIELD,array($field->name,$field->orgTable),E_USER_WARNING);
            continue;
          }
          if($res) {
            $names .= ', ';
            $values .= ', ';
          } else {
            $res = true;
          }
          $names .= $this->connection->quoteFieldName($field->orgName);
          $values .= $field->value->getAsSqlQuoted();
        }
        if(!$res) {
          throw new EException(E_CANNOT_INSERT_BLANK_ROW);
        }
        $res = $this->connection->execute("INSERT INTO "
          .$this->connection->quoteFieldName($this->table)
          ." (".$names.") VALUES ("
          .$values.")"
        );
        if($res) {
          foreach($this->fields as /** @var TField */ $field) {
            if($field->autoValue) {
              $field->value->setRaw($this->connection->getInsertId());
            }
          }
          $this->inserting = false;
        }
        return $res;
      } else {
        $query = '';
        $res = false;
        foreach($this->fields as /** @var TField */ $field) {
          if(!$field->hasChanged()) {
            continue;
          }
          if($field->isNull()) {
            if($field->notNull) {
              triggerError(E_FIELD_CANNOT_BE_NULL,array($field->name),E_USER_WARNING);
            }
          }
          if($field->orgTable != $this->table) {
            triggerError(E_CANNOT_UPDATE_JOINED_TABLE_FIELD,array($field->name,$field->orgTable),E_USER_WARNING);
            continue;
          }          
          if($res) {
            $query .= ', ';
          } else {
            $res = true;
          }
          $query .= $this->connection->quoteFieldName($field->orgName)." = ".$field->value->getAsSqlQuoted();
        }

        if(!$res) {
          return true;
        }

        $res = $this->connection->execute("UPDATE "
          .$this->connection->quoteFieldName($this->table)
          ." SET ".$query." WHERE "
          .$this->getPrimaryKeyCondition()
        );
        if($res) {
          foreach($this->fields as /** @var TField */ $field) {
            $field->setOldValue($field->value->getAsSql());
          }
        }
        return $res;
      }
    }
    /**
    * Fetch row values as an associative array and move to next row
    *
    * @return mixed[]
    */
    public function fetchAssoc() {
      $res = $this->doFetchAssoc();
      if($res) {
        $this->rowIndex ++;
      }
      return $res;
    }
    /**
    * Fetch value of one single field and movet to next row
    *
    * @param mixed Field name or field index
    * @return mixed
    */
    public function fetchValue($fieldNameOrIndex) {
      $fieldNameOrIndex = $this->fieldNameToIndex($fieldNameOrIndex);
      $res = $this->fetchArray();
      if($res) {
        $res = $res[$fieldNameOrIndex];
      }
      return $res;
    }
    /**
    * Fetch all rows into single dimension array of scalar values
    *
    * @param string Field that's values should be fetched 
    * @return mixed[]
    */
    public function fetchValueList($fieldNameOrIndex)  {
      $fieldNameOrIndex = $this->fieldNameToIndex($fieldNameOrIndex);
      $this->first();
      $res = array();
      while($val = $this->fetchValue($fieldNameOrIndex)) {
        $res[] = $val;
      }
      return $res;
    }
    /**
    * Fetch all rows into single dimension associative array of scalar values
    *
    * @param string Field that's values should be fetched 
    * @param string Field that's values should be the keys of result array
    * @return mixed[]
    */
    public function fetchValueMap ($valueField, $keyField)  {
      $valueField = $this->fieldNameToIndex($valueField);
      $keyField = $this->fieldNameToIndex($keyField);
      $res = array();
      while($row = $this->fetchArray()) {
        $res[$row[$keyField]] = $row[$valueField];
      }
      return $res;
    }
    /**
    * Fetch all rows into array of objects
    *
    * @return obj[]
    */
    public function fetchObjectList()  {
      $res = array();
      while($obj = $this->fetchObject()) {
        $res[] = $obj;
      }
      return $res;
    }
    /**
    * Fetch all rows into an associative array of objects
    * 
    * @param string Field that's values should be the keys of result array
    * @return mixed[]
    */
    public function fetchObjectMap ($keyField)  {
      $res = array();
      while($obj = $this->fetchObject()) {
        $res[$obj->$keyField] = $obj;
      }
      return $res;
    }
    /**
    * Fetch all rows into an associative array of array of objects
    * 
    * @param string Field that's values should be the keys of result array
    * @return mixed[]
    */
    public function fetchObjectGroups($groupField, $textComparsion = false)  {
      $res = array();
      while($obj = $this->fetchObject()) {
        if($textComparsion) {
          $x = upperCase(trim($obj->$groupField));
        } else {
          $x = $obj->$groupField;
        }
        $res[$x][] = $obj;
      }
      return $res;
    }    
    /**
    * Fetch tree of objects, where chlid rows are stored inside "childRows"
    * property in an associative array
    *
    * @param string Field that uniquely identifies a row
    * @param string Field that points to a parent row
    * @param string Field that's values should be the keys of "childRows" arrays 
    * @return obj[]
    */
    public function fetchObjectTreeMap($idField, $parentField, $keyField, $backReference = false)  {
         
      $lst = $this->fetchObjectList();
      $root = array();      
      foreach($lst as $obj) {
        $is_root = true;
        foreach ($lst as $p_obj) {
          
          if($obj->$parentField === $p_obj->$idField) {
            $is_root = false;
            if($backReference) {
              $obj->parentRow = $p_obj;
            }
            if($keyField) {
              $p_obj->childRows[$obj->$keyField] = $obj;
            } else {
              $p_obj->childRows[] = $obj;
            }
          }
        }
        if($is_root) {
          if($keyField) {
            $root[$obj->$keyField] = $obj;
          } else {
            $root[] = $obj;
          }
        }
      }
      return $root;    
    }
    /**
    * Fetch tree of objects, where chlid rows are stored inside "childRows"
    * property in an enumerated array
    *
    * @param string Field that uniquely identifies a row
    * @param string Field that points to a parent row 
    * @return obj[]
    */
    public function fetchObjectTree($idField, $parentField, $backReference = false)  {

      $lst = $this->fetchObjectList();
      $root = array();      
      foreach($lst as $obj) {
        $is_root = true;
        foreach ($lst as $p_obj) {
          if($obj->$parentField === $p_obj->$idField) {
            $is_root = false;
            if($backReference) {
              $obj->parentRow = $p_obj;
            }
            $p_obj->childRows[] = $obj;
          }
        }
        if($is_root) {
          $root[] = $obj;
        }
      }
      return $root;    
    }
  }

 

  /**
  *
  */
  abstract class TCustomConnection extends TObject {
    /** @var TClientInfo Client library information */
    public $client;
    private $connected;
    /** @var bool Throw exceptions? */
    public $debug = false;
    /** @var TDataSet Result of the query */
    public $result = null;
    /** @var bool Does the server support transactions? */
    protected $transactionsSupported = false;
    /** @var bool Does the server contain multiple databases? */
    protected $multiDatabase = false;        
    private $lastQuery;
    private $lastErrorMessage;
    private $lastErrorCode;
    private $transactionLevel = 0;
    
    protected function setLastError($code,$message) {
      $this->lastErrorCode = $code;
      if(is_int($message)) {
        $message = formatError($message);        
      }
      $this->lastErrorMessage = $message;      
    }
    /**
    * Set enconding in which the output should be returned 
    *
    * @param string Encoding name according to ICONV function
    * @return bool Success
    */
    abstract protected function setOutputEncoding($encoding);
    /**
    * Returns number of rows matched by last WHERE query 
    *
    * @return int
    */
    abstract public function getMatchedCount();
    /**
    * Returns number of rows affected by last query 
    *
    * @return int
    */
    abstract public function getAffectedCount();
    private $database;
    /**
    * Get last executed query
    *
    * @return string
    */
    public function getLastQuery( )  {
      return $this->lastQuery;
    }            
    abstract protected function doConnect();
    abstract protected function doQuery($query, $fetchResult);
    /**
    * Get last auto-value generated while inserting new row
    *
    * @return mixed
    */
    abstract public function getInsertId();
    /**
    * Is the connection established?
    *
    * @return bool
    */
    public function isConnected() {
      return $this->connected;
    }
    abstract protected function doClose();
    /**
    * Close connection
    *
    * @return bool
    */
    public function close( )  {
      $res = $this->doClose();
      if(!$res) {
        if($this->debug) {
          throw new EException(E_CONNECTION_COULD_NOT_BE_CLOSED,array($this->lastErrorCode,$this->lastErrorMessage));
        } else {
          triggerError(E_CONNECTION_COULD_NOT_BE_CLOSED,array($this->lastErrorCode,$this->lastErrorMessage),E_USER_WARNING);
        }
      }
      if($res) {
        $this->connected = false;
      }
      return $res;
    }
    /**
    * Check if the server is connected
    *
    * @return bool
    */
    protected function checkConnected() {
      $res = $this->isConnected();
      if(!$res) {
        $this->setLastError(0,E_NOT_CONNECTED_TO_DATABASE_SERVER);
      }
      return $res; 
    }
    private function execQuery($query, $fetchResult) {      
      $res = $this->checkConnected();
      if($res) {
        $this->lastQuery = $query;
        $res = $this->doQuery($query,$fetchResult);
      }
      if(!$res) {
        if($this->debug) {
          throw new EQueryException(E_COULD_NOT_EXECUTE_QUERY,array($this->lastErrorCode,$this->lastErrorMessage,$query));
        } else {
          triggerError(E_COULD_NOT_EXECUTE_QUERY,array($this->lastErrorCode,$this->lastErrorMessage,$query),E_USER_WARNING);
        }
      }        

      return $res;
 
    }
    protected function doBegin() {
      return $this->doQuery('START TRANSACTION',false);
    }
    protected function doCommit() {
      return $this->doQuery('COMMIT',false);
    }
    protected function doRollback() {
      return $this->doQuery('ROLLBACK',false);
    }
    /**
    * Get textual information about client library
    *
    * @return string
    */
    public function getClientInfo() {
      return $this->doGetClientInfo();
    }
    /**
    * Get version number of the client library
    *
    * @return string
    */
    public function getClientVersion() {
      return $this->doGetClientVersion();
    }
    /**
    * Get textual information about server
    *
    * @return string
    */
    public function getServerInfo() {
      if($this->checkConnected()) {
        return $this->doGetServerInfo();
      } else {
        return false;
      }
    }
    /**
    * Get version number of the server
    *
    * @return string
    */
    public function getServerVersion() {
      if($this->checkConnected()) {
        return $this->doGetServerVersion();
      } else {
        return false;
      }
    }
    
    protected function doGetClientInfo() {
      return 'N/A';
    }
    protected function doGetClientVersion() {
      return 0;
    }
    protected function doGetServerInfo() {
      return 'N/A';
    }
    protected function doGetServerVersion() {
      return 0;
    }
    /**
    * Connect to the database server
    *
    * @return bool
    */
    public function connect() {
      $res = $this->doConnect();
      if($res) {
        $this->connected = true;
      } elseif($this->debug) {
        throw new EException(E_COULD_NOT_CONNECT_TO_DATABASE_SERVER,array($this->lastErrorCode,$this->lastErrorMessage));
      } else {
        triggerError(E_COULD_NOT_CONNECT_TO_DATABASE_SERVER,array($this->lastErrorCode,$this->lastErrorMessage),E_USER_WARNING);
      }
      if(!$this->setOutputEncoding(CFG_ENCODING)) {
        triggerError(E_COULD_NOT_SET_DB_ENCODING,array($this->lastErrorCode,$this->lastErrorMessage),E_USER_WARNING);
      }
    }
    private function transOperation($operation) {
      if($this->transactionsSupported) {
        $res = $this->{'do'.$operation}();
      } else {
        $this->setLastError(0,E_TRANSACTIONS_NOT_SUPPORTED);
        $res = true;
      }
      if(!$res) {
        if($this->debug) {
          throw new EException(E_TRANSACTION_ERROR,array($this->lastErrorCode,$this->lastErrorMessage));
        } else {
          triggerError(E_TRANSACTION_ERROR,array($this->lastErrorCode,$this->lastErrorMessage),E_WARNING);
        }
      }
      return $res;
    }
    /**
    * Begin transaction
    *
    * @return bool
    */
    public function begin() {
      $res = true;
      if($this->transactionLevel == 0) {
        $res = $this->transOperation('begin'); 
      }
      if($res) {
        $this->transactionLevel ++;
      }
      return $res;
    }
    /**
    * Commit transaction
    *
    * @return bool
    */
    public function commit() {
      $res = true;
      if($this->transactionLevel == 1) {
        $res = $this->transOperation('commit'); 
      }
      if($res && ($this->transactionLevel > 0)) {
        $this->transactionLevel --;
      }
      return $res;
    }
    /**
    * Rollback transaction
    *
    * @return bool
    */
    public function rollback() {
      $res = true;
      if($this->transactionLevel == 1) {
        $res = $this->transOperation('rollback'); 
      }
      if($res && ($this->transactionLevel > 0)) {
        $this->transactionLevel --;
      }
      return $res; 
    }
    /**
    * Executes query and fetches the result
    *
    * @return TDataSet
    */
    public function query($query) {
      $res = $this->execQuery($query,true);
      if($res) {
        return $this->result;
      } else {
        false;
      }
    }
    /**
    * Escape the value to be used in SQL statements 
    *
    * @return string
    */
    public function escape($value) {
      return addslashes($value);
    }
    /**
    * Escape the value to be used in SQL statements and enquotes it 
    *
    * @return string
    */
    public function quote($value) {
      return "'".$this->escape($value)."'";
    }
    /**
    * Escape the field name to be used in SQL statements 
    *
    * @return string
    */
    public function escapeFieldName($value) {
      return $value;
    }
    /**
    * Escape the field name to be used in SQL statements and enquotes it
    *
    * @return string
    */
    public function quoteFieldName($value) {
      return '`'.$value.'`';
    }
    /**
    * Is the driver in transaction mode 
    *
    * @return bool
    */
    public function isTransactionPending ( )  {
      return ($this->transactionLevel > 0);
    }
    /**
    * Executes query that doesn't return any result
    *
    * @return bool
    */
    public function execute($query)  {
      return ($this->execQuery($query,false) && true);
    }
    protected function doSelectDatabase($databaseName) {
      return $this->doQuery('USE '.$databaseName,false);
    }
    /**
    * Select database 
    *
    * @return bool
    */
    public function selectDatabase($databaseName) {
      if($this->multiDatabase) {
        $res = $this->doSelectDatabase($databaseName);
      } else {
        $res = false;
        $this->lastErrorCode = 0;
        $this->lastErrorMessage = formatError(E_SERVER_IS_NOT_MULTIDATABASE);
      } 
      if($res) {
        $this->database = $databaseName;        
      }
      if(!$res) {
        if($this->debug) {
          throw new EException(E_COULD_NOT_SELECT_DATABASE,array($this->lastErrorCode,$this->lastErrorMessage));
        } else {
          triggerError(E_COULD_NOT_SELECT_DATABASE,array($this->lastErrorCode,$this->lastErrorMessage),E_USER_WARNING);
        }
      }
    }
    
    /**
    * Returns the current database name
    *
    * @return string
    */
    public function getCurrentDatabase()  {
      return $this->database;     
    }
    public function __construct() {

    }
  }

  class TSqlConnectionParameters extends TObject {
    /** @var string Host name of the MySQL server */
    public $host = '127.0.0.1';
    /** @var int Port on which the MySQL server listens */
    public $port = 3306;
    /** @var string Login of the user to access the server */
    public $username = null;
    /** @var string Password for the user */
    public $password = null;
    
    /** @var string Textual information about the server */
    public $info = null;
    /** @var int Version of the server */
    public $version = null;
  }

  $__roadDataTypeValueMatrix = array(
    TYPE_BOOLEAN => 'TBoolean',
    TYPE_DATE => 'TDate',
    TYPE_DATETIME => 'TDateTime',
    TYPE_TIME => 'TTime',
    TYPE_INTEGER => 'TNumber',
    TYPE_DECIMAL => 'TNumber',
    TYPE_STRING => 'TText',
    TYPE_MEMO => 'TText',
    TYPE_SET => 'TValue',
    TYPE_ENUM => 'TValue',
    TYPE_UNKNOWN => 'TValue'    
  );
  $__roadDataTypeStringMatrix = array(
    TYPE_BOOLEAN => 'boolean',
    TYPE_DATE => 'date',
    TYPE_DATETIME => 'dateTime',
    TYPE_TIME => 'time',
    TYPE_INTEGER => 'integer',
    TYPE_DECIMAL => 'decimal',
    TYPE_STRING => 'string',
    TYPE_MEMO => 'memo',
    TYPE_SET => 'set',
    TYPE_ENUM => 'enum',
    TYPE_UNKNOWN => 'unknown'    
  );
  /**
  * Returns the name of provided data type
  *
  * @param int Data type
  * @return string
  */
  function dataTypeToStr($dataType)  {
    global $__roadDataTypeStringMatrix;
    if(!isset($__roadDataTypeStringMatrix[$dataType])) {
      $dataType = TYPE_UNKNOWN;
    }
    return $__roadDataTypeStringMatrix[$dataType];
  }
  /**
  * Outputs info about provided fields
  *
  * @param TField[]
  * @return void
  */
  function fieldInfo($fields)  {
    echo "<ul>\n";
    foreach($fields as /** @var TField */ $field) {
      echo '<li>'.($field->primaryKey ? '<u>' : '').$field->name.($field->primaryKey ? '</u>' : '').' '.dataTypeToStr($field->dataType)." (".$field->length.")".($field->notNull ? ' not null' : '').($field->autoValue ? ' auto_increment' : '').'  =  '.$field->value->getAsSqlQuoted()."</li>\n";
    }
    echo "</ul>\n";      
  }  

  
?>